#include "libDM_icm42688.hpp"
#include "libDM_mmc5983ma.hpp"
#include "libMM_lps22hb.hpp"

#include <array>
#include <iterator>

constexpr uint32_t MS_TO_US = 1000U;

// Common declarations
struct PIN_CONFIG {
  uint8_t pinSS = 15;
  uint8_t pinMmcSS = 16;
  uint8_t pinLpsSS = 17;
  uint8_t pinMosi = 12U;
  uint8_t pinMiso = 13U;
  uint8_t pinClk = 14U;

  uint8_t pinSdioClk = 38U;
  uint8_t pinSdioDat0 = 39U;
  uint8_t pinSdioDat1 = 42U;
  uint8_t pinSdioDat2 = 21U;
  uint8_t pinSdioDat3 = 41U;
  uint8_t pinSdioCmd = 40U;

  uint8_t pinGpsRx = 8;
  uint8_t pinGpsTx = 18;

  uint8_t pinVn200Rx = 17;
  uint8_t pinVn200Tx = 16;

  uint8_t pinProtocolRx = 5;
  uint8_t pinProtocolTx = 4;
};

extern const PIN_CONFIG pinConf;

// ====================================================== //
// =================== Configurations =================== //
// ====================================================== //
// Place here common configurations that are used in multiple files.

/**
   @brief Configuration parameters for the application.
*/
struct GENERAL_CONFIG {
  uint32_t sdioClock = 40000U;  // [kHz]
  const char* sdioMountpoint = "/sdcard";

  uint32_t icmSpiClock = 24000000U;  // [Hz]

  uint32_t lpsSpiClock = 10000000U;  // [Hz]

  uint32_t mmcSpiClock = 10000000U;  // [Hz]

  uint32_t vn200Baudrate = 921600U;  // [bps]

  uint32_t gpsBaudrate = 230400U;  // [bps]
};

extern const GENERAL_CONFIG generalConf;

// ====================================================== //
// ================ Sensors configuration =============== //
// ====================================================== //
// NOLINTBEGIN(misc-non-private-member-variables-in-classes)

/**
   @class SENSOR_UNIT_CONFIG
   @brief This class represents the configuration for a sensor.
*/
class SENSOR_UNIT_CONFIG {
public:
  bool enable = true;
  uint32_t period = 2U;  // [ms]
  uint32_t ratioToTask =
    1U;                            // [] how many ratio of the current loop are necessary to require sensor parsing
  uint32_t timestampPrev = 0U;     // [us]
  uint32_t taskTicksCounter = 0U;  // [] counts the number of task ticks since last sensor parsing

  /**
       @brief Constructs a new SENSOR_UNIT_CONFIG object with the specified enable and period
       values.
       @param enable Whether the sensor unit task is enabled or not.
       @param period The period of the sensor unit task in milliseconds.
    */
  SENSOR_UNIT_CONFIG(bool enable, uint32_t period)
    : enable(enable), period(period) {}

  /**
       @brief Sets the ratio of the current loop necessary to require sensor parsing.
       @param taskPeriod The period of the current loop in milliseconds.
       @return True if the ratio was set successfully, false otherwise.
    */
  bool setRatio(uint32_t taskPeriod) {
    ratioToTask = 1U;

    if (taskPeriod <= period) {
      ratioToTask = period / taskPeriod;
      return true;  // NOLINT(readability-simplify-boolean-expr)
    }

    return false;
  }

  /**
       @brief Sets the period of the sensor unit task.
       @param newPeriod The new period of the sensor unit task in milliseconds.
    */
  void setPeriod(uint32_t newPeriod) {
    period = newPeriod;
  }

  /**
       @brief Checks if the sensor has to be parsed in the current loop.
       @return True if the sensor has to be parsed, false otherwise.
    */
  bool checkRatio() {
    ++taskTicksCounter;
    if (taskTicksCounter >= ratioToTask)
      return true;  // NOLINT(readability-simplify-boolean-expr)

    return false;
  }

  /**
       @brief Resets the task ticks counter.
    */
  void resetRatioCounter() {
    taskTicksCounter = 0U;
  }
};

// ====================================================== //
// ======================= Objects ====================== //
// ====================================================== //
ICM42688_SPI icm(generalConf.icmSpiClock,
                 pinConf.pinSS,
                 &SPI,
                 ABSTRACT_SENSOR_SPI::SPI_CONF(),
                 true,
                 true);
MMC5983MA_SPI mmc(generalConf.mmcSpiClock, pinConf.pinMmcSS, &SPI, ABSTRACT_SENSOR_SPI::SPI_CONF());
LPS22HB_SPI lps(generalConf.lpsSpiClock,
                pinConf.pinLpsSS,
                false,
                &SPI,
                ABSTRACT_SENSOR_SPI::SPI_CONF());

ABSTRACT_IMU::IMU_OUTPUT resIMU;
ABSTRACT_MAG::MAG_OUTPUT resMAG;
ABSTRACT_BARO::BARO_OUTPUT resBARO;

// ====================================================== //
// ================== Config structures ================= //
// ====================================================== //

const PIN_CONFIG pinConf = PIN_CONFIG();              // NOLINT
const GENERAL_CONFIG generalConf = GENERAL_CONFIG();  // NOLINT

/**
   @brief Configuration structure for sensor units.
*/
struct
{
  SENSOR_UNIT_CONFIG sensorIMU = SENSOR_UNIT_CONFIG(true, 2);    // 500Hz so 2ms
  SENSOR_UNIT_CONFIG sensorMag = SENSOR_UNIT_CONFIG(true, 10);   // 100Hz so 10ms
  SENSOR_UNIT_CONFIG sensorBaro = SENSOR_UNIT_CONFIG(true, 14);  // 71.4Hz so 14ms
} SENSORS_CONFIG;

// The sensors polling frequency will be adjusted considering 500Hz task frequency. This frequency
// can be increased with no limit as long as your cpu can keep up!
uint32_t taskPeriod = 2;  // [ms] 2ms <-> 500Hz

uint32_t lastTime = 0U;

void pinInit();       // Defined below;
void protocolInit();  // Defined below;
void initSensors();   // Defined below;

// the setup function runs once when you press reset or power the board
void setup() {
  Serial.begin(250000);

  pinInit();
  protocolInit();
  initSensors();
}

// the loop function runs over and over again forever
void loop() {
  if ((micros() - lastTime) >= taskPeriod * MS_TO_US)
  {
    lastTime = micros();

    // ~~~~~~~~~~~~~~~~~ ICM ~~~~~~~~~~~~~~~~~ //
    // IMU will be read at task speed so no need to check ratio
    icm.readFifo();
    resIMU = icm.getOutput();

    // ~~~~~~~~~~~~~~~~~ MMC ~~~~~~~~~~~~~~~~~ //
    if (SENSORS_CONFIG.sensorMag.checkRatio())
    {
      mmc.readMag();
      resMAG = mmc.getOutput();
      if (resMAG.usMagTimestamp != SENSORS_CONFIG.sensorMag.timestampPrev)
      {        
        SENSORS_CONFIG.sensorMag.timestampPrev = resMAG.usMagTimestamp;
        SENSORS_CONFIG.sensorMag.resetRatioCounter();
      }
    }

    // ~~~~~~~~~~~~~~~ LPS22HB ~~~~~~~~~~~~~~~ //
    if (SENSORS_CONFIG.sensorBaro.checkRatio())
    {
      lps.readTemperaturePressureAltitude();
      resBARO = lps.getOutput();

      if (resBARO.usPressureTimestamp != SENSORS_CONFIG.sensorBaro.timestampPrev)
      {
        SENSORS_CONFIG.sensorBaro.timestampPrev = resBARO.usPressureTimestamp;
        SENSORS_CONFIG.sensorBaro.resetRatioCounter();
      }
    }
  }
}

void pinInit() {
  pinMode(pinConf.pinSS, OUTPUT);
  pinMode(pinConf.pinLpsSS, OUTPUT);
  pinMode(pinConf.pinMmcSS, OUTPUT);
  digitalWrite(pinConf.pinLpsSS, HIGH);
  digitalWrite(pinConf.pinSS, HIGH);
  digitalWrite(pinConf.pinMmcSS, HIGH);
}

void protocolInit() {
  // SPI init
#if defined(ESP_PLATFORM)
  SPI.begin(pinConf.pinClk, pinConf.pinMiso, pinConf.pinMosi, pinConf.pinSS);
#else
  SPI.begin();
#endif
  delay(1);
}

void initSensors()
{
  // ====================================================== //
  // ====================== Init ICM ====================== //
  // ====================================================== //
  if (SENSORS_CONFIG.sensorIMU.enable)
  {
    icm.softResetSensor();
    icm.begin();
    icm.setGyroOdrRange(ICM42688_ENUM::ODR::_1kHz, ICM42688_ENUM::GYRO_RANGE::_2000dps);
    SENSORS_CONFIG.sensorIMU.setPeriod(SENSORS_CONFIG.sensorIMU.period); // IMU period = 2ms
    SENSORS_CONFIG.sensorIMU.setRatio(taskPeriod);                       // Task period = 2ms
    Serial.println("Main task: IMU task ratio = "
                   + String(SENSORS_CONFIG.sensorIMU.ratioToTask));
    delay(1);
    icm.setAccOdrRange(ICM42688_ENUM::ODR::_1kHz, ICM42688_ENUM::ACC_RANGE::_16g);
    icm.setGyroAccelFilter(ICM42688_ENUM::FILTER_BW::ODR_DIV_4,
                           ICM42688_ENUM::FILTER_BW::ODR_DIV_4,
                           ICM42688_ENUM::FILTER_ORDER::_2nd_ORDER,
                           ICM42688_ENUM::FILTER_ORDER::_2nd_ORDER);
    delay(100);                                     // NOLINT
    if (icm.checkImmobility(0.3F, 200U, 1U, 1000U)) // NOLINT
      icm.calibGyro(2000);                        // NOLINT
    icm.flushFifo(true);
    delay(100); // NOLINT
  }

  // ====================================================== //
  // ====================== Init MMC ====================== //
  // ====================================================== //
  if (SENSORS_CONFIG.sensorMag.enable)
  {
    mmc.softResetSensor();
    delay(10); // NOLINT
    mmc.begin();
    mmc.setMagOdrBw(MMC5983MA_ENUM::ODR::_100HZ, MMC5983MA_ENUM::BW::_200HZ);
    SENSORS_CONFIG.sensorMag.setPeriod(SENSORS_CONFIG.sensorMag.period); // MAG period = 10ms
    SENSORS_CONFIG.sensorMag.setRatio(taskPeriod);                       // Task period = 2ms
    Serial.println("Main task: MAG task ratio = "
                   + String(SENSORS_CONFIG.sensorMag.ratioToTask));
    mmc.setAutoSetResetState(true);
  }

  // ====================================================== //
  // ====================== Init LPS ====================== //
  // ====================================================== //
  if (SENSORS_CONFIG.sensorBaro.enable)
  {
    lps.begin();
    lps.setOdr(LPS22HB_ENUM::ODR::_75HZ);
    delay(100);                                                            // NOLINT
    SENSORS_CONFIG.sensorBaro.setPeriod(SENSORS_CONFIG.sensorBaro.period); // BARO period = 14ms
    SENSORS_CONFIG.sensorBaro.setRatio(taskPeriod);                        // Task period = 2ms
    Serial.println("Main task: BARO task ratio = "
                   + String(SENSORS_CONFIG.sensorBaro.ratioToTask));
    // lps.setLowPassFilterState(LPS22HB_ENUM::FILTER_BW::ODR_DIV_20);
    lps.configureReferenceParameters(
      ABSTRACT_BARO_ENUM::BARO_VALUES::INVALID_PRESSURE,
      ABSTRACT_BARO_ENUM::BARO_VALUES::DEFAULT_SEA_LEVEL_TEMPERATURE,
      ABSTRACT_BARO_ENUM::BARO_VALUES::INVALID_ALTITUDE);
  }
}
